library(rsimsum)

# The functions used to fit the models are in fit_univariate_models.R 
# The functions used to calculate mean CRPS are in CRPS.R

set.seed(1729) # Do not change this. 

n_sim = 2000 # Number of repeated simulations
n_obs = 300 # Sample size of datasets
n_sen = 16 # Number of different scenarios
n_met = 4 # Number of methods tested
n_par = 5+4 # Number of estimated parameters(regression parm. + calculated parm.)
n_test = 300 # Size of test set for calculating CRPS
alpha = 0.05 # For tests and confidence intervals

methods = c("TP", "TPIC", "Substitute", "Tobit") # List of methods
par_name = c("gamma0", "gamma1", "sigma", 
             "beta0", "beta1",
             "mcrps", "marg_mean0", "marg_mean1", "marg_eff_obs") # List of parameter names

# Data frames for storing results:
summ.all = data.frame(stat = character(), est = numeric(), mcse = numeric(), method = character(), 
                      LOD = character(), disc_prob = character(), beta1 = character(),
                      par = character(), lower = numeric(), upper = numeric())

data.all = data.frame(dataset = integer(), method = character(), theta = numeric(), se = numeric(),
                      par = character(), LOD = character(), disc_prob = character(), .dropbig = logical())

res.list = list()

# Varying factors:
LOD_real = rep(c(c(0.5,1,2,3)), each = 4)
LOD_vec = log(LOD_real) # Limit of Detection 
LOD_n = gsub(".", ",", as.character(LOD_real), fixed = TRUE)


beta0_vec = rep(c(-1.2,-0.7,-0.5,-0.2), 4)
disc_prob = rep(c(10, 20, 30, 40), 4)
dp_n = gsub(".", ",", as.character(disc_prob), fixed = TRUE)

# Constant factors:
proportion_diagnosed = 0.5
beta1 = -0.2

gamma0 = 1 # (single: 1, multiple: 0.8)
gamma1 = 0.2 # continuous part

sigma = 0.7 # continuous part

true.crps = numeric(n_sen) # Compute CRPS of true distribution

# Store failed data for analysis
xfail = c()
yfail = c()
dfail = c()
LODfail = c()
beta0fail = c()

### ITERATE OVER SCENARIOS:
for(sen in 1:n_sen){

  pos_res_sen = 1 
  res.simsum.sen = data.frame(dataset = rep(NA, n_sim*n_met*n_par), # Scenario number
                              method = rep(NA, n_sim*n_met*n_par), # Name of the method
                              theta = rep(NA, n_sim*n_met*n_par), # Parameter estimate
                              se = rep(NA, n_sim*n_met*n_par), # Estimated standard error
                              par = rep(NA, n_sim*n_met*n_par), # Name of parameter
                              LOD = rep(NA, n_sim*n_met*n_par), # Limit of detection for dataset
                              disc_prob = rep(NA, n_sim*n_met*n_par) ) # Discrete probability for dataset
  
  # Varying bet0
  beta0 = beta0_vec[sen]
  
  # Varying the LOD
  LOD = LOD_vec[sen]
  LOD_name = LOD_n[sen]
  
  # Calculate true marginal means and effects:
  Z_0 = (LOD - gamma0)/sigma
  C_0 = (1-pnorm(Z_0 - sigma))
  Z_1 = (LOD - gamma0 - gamma1)/sigma
  C_1 = (1-pnorm(Z_1 - sigma))
  marginal_effect = (1-pnorm(beta0 + beta1))/(1-pnorm(beta0)) * C_1 / C_0 * exp(gamma1)
  marginal_mean0 = (1-pnorm(beta0)) * (1-pnorm(Z_0 - sigma)) * exp(gamma0 + sigma^2/2)
  marginal_mean1 = (1-pnorm(beta0 + beta1)) * (1-pnorm(Z_1 - sigma)) * exp(gamma0 + gamma1 + sigma^2/2)

  ### SIMULATIONS:
  for(i in 1:n_sim){
    
    beta = c(beta0, beta1)
    gamma = c(gamma0, gamma1)
    
    # TRAINING DATA:
    x1 = rbinom(n = n_obs, size = 1, p = proportion_diagnosed) #"diagnosis"
    #x2 = rnorm(n = n_obs, mean = 0, sd = 1)
    #x3 = rpois(n = n_obs, lambda = 1)
    x = cbind(1, x1)
    
    p = pnorm(x %*% beta)
    discrete_part = as.integer(runif(n_obs) < p)
    
    # On log-scale:
    errors = rnorm(n = n_obs, mean = 0, sd = sigma)
    y = x %*% gamma + errors
    mean(y <= LOD)
    y[y <= LOD] = NA # censored
    y[discrete_part == 1] = NA # true zeroes
    
    # TEST DATA FOR CRPS:
    x10 = rbinom(n = n_test, size = 1, p = proportion_diagnosed) #"diagnosis"
    #x20 = rnorm(n = n_test, mean = 0, sd = 1)
    #x30 = rpois(n = n_test, lambda = 1)
    x0 = cbind(1, x10)

    p0 = pnorm(x0 %*% beta)
    discrete_part0 = as.integer(runif(n_test) < p0)

    # On real-scale:
    errors = rnorm(n = n_test, mean = 0, sd = sigma)
    y0_log = x0 %*% gamma + errors
    y0_log[y0_log <= LOD] = NA # censored
    y0_log[discrete_part0 == 1] = NA # true zeroes
    y0 = exp(y0_log)
    y0[is.na(y0)] = 0

    # Likelihood maximization:
    par = c(gamma, sigma, beta)
    solution_ni = fit_model(x = x, y = y, LOD = LOD, model = "ni", par)
    solution_i = fit_model(x = x, y = y, LOD = LOD, model = "i", par)
    solution_pi = fit_model(x = x, y = y, LOD = LOD, model = "pi", par)
    par = c(solution_pi$par[1:(length(gamma)+1)], beta)
    solution_t = fit_model(x = x, y = y, LOD = LOD, model = "t", par)
    
    solutions = list(solution_ni, solution_i, solution_pi, solution_t)
    
    # Theoretically optimal CRPS from true distribution: 
    tc = mean.CRPS(y0, x0, m = 1, gamma_est = gamma, sigma_est = sigma, beta_est = beta, realLOD = exp(LOD))
    true.crps[sen] = true.crps[sen] + 1/n_sim * tc
    
    # Storing results:
    for(met in 1:n_met){
      #Checking for errors:
      if(solutions[[met]]$convergence != 0 | 
         class(try(solve(-solutions[[met]]$hessian), silent = TRUE)) != "matrix"){
        # Non-convergence
        print("Failed solution")
        print(methods[met])
        print(sen)
        res.simsum.sen[pos_res_sen:(pos_res_sen+n_par-1),] = cbind( rep(sen,n_par), 
                                                                    rep(methods[met],n_par),
                                                                    rep(NA, n_par), rep(NA, n_par), par_name,
                                                                    rep(LOD_name,n_par), rep(dp_n[sen],n_par), FALSE) 
        pos_res_sen = pos_res_sen + n_par
        
        # Store failed data for analysis
        xfail = rbind(xfail, x1)
        yfail = rbind(yfail, t(y))
        dfail = rbind(dfail, discrete_part)
        LODfail = c(LODfail, LOD)
        beta0fail = c(beta0fail, beta0)
        
      }else if(sum(diag(solve(-solutions[[met]]$hessian)) < 0) > 0){
        # Negative standard errors
        print("Failed SE")
        print(methods[met])
        print(sen)
        res.simsum.sen[pos_res_sen:(pos_res_sen+n_par-1),] = cbind( rep(sen,n_par),
                                                                    rep(methods[met],n_par),
                                                                    rep(NA, n_par), rep(NA, n_par), par_name,
                                                                    rep(LOD_name,n_par), rep(dp_n[sen],n_par), FALSE) 
        pos_res_sen = pos_res_sen + n_par
        
        # Store failed data for analysis
        xfail = rbind(xfail, x1)
        yfail = rbind(yfail, t(y))
        dfail = rbind(dfail, discrete_part)
        LODfail = c(LODfail, LOD)
        beta0fail = c(beta0fail, beta0)
        
      }else{
        # Successfull solution
        
        C = solve(-solutions[[met]]$hessian) # Inverse negative Hessian
        SE = sqrt(diag(C)) # Standard errors
        solutions[[met]]$par[length(gamma)+1] = abs(solutions[[met]]$par[length(gamma)+1]) # Ensure right sign
        
        gamma.res = solutions[[met]]$par[1:length(gamma)]
        sigma.res = solutions[[met]]$par[length(gamma)+1]
        
        if(length(SE) == length(gamma) + length(beta) + 1){ # Two-part models
          beta.res = solutions[[met]]$par[(length(gamma)+2):(2*length(gamma)+1)]
          
          # Calculate marginal effects and CPRS
          if(methods[met] == "TP"){
            marg_eff = (1 - pnorm(beta.res[1]+beta.res[2])) / (1 - pnorm(beta.res[1])) * exp(gamma.res[2])
            marg_mean0 = (1 - pnorm(beta.res[1])) * exp(gamma.res[1] + sigma.res^2/2) 
            marg_mean1 = (1 - pnorm(beta.res[1] + beta.res[2])) * exp(gamma.res[1] + gamma.res[2] + sigma.res^2/2) 
        
            mcrps = mean.CRPS(y0 = y0, x0 = x0, m = 2, gamma_est = gamma.res,
                               sigma_est=sigma.res, beta_est = beta.res, realLOD = exp(LOD))
             
          }else if(methods[met] == "TPIC"){
            Z_0 = (LOD - gamma.res[1])/sigma.res
            C_0 = (1-pnorm(Z_0 - sigma.res))
            Z_1 = (LOD - gamma.res[1] - gamma.res[2])/sigma.res
            C_1 = (1-pnorm(Z_1 - sigma.res))
            marg_eff = (1-pnorm(beta.res[1] + beta.res[2]))/(1-pnorm(beta.res[1])) * C_1 / C_0 * exp(gamma.res[2])
            marg_mean0 = (1-pnorm(beta.res[1])) * (1-pnorm(Z_0 - sigma.res)) * exp(gamma.res[1] + sigma.res^2/2)
            marg_mean1 = (1-pnorm(beta.res[1] + beta.res[2])) * (1-pnorm(Z_1 - sigma.res)) * exp(gamma.res[1] + gamma.res[2] + sigma.res^2/2)
            
            mcrps = mean.CRPS(y0, x0, m = 1, gamma_est = gamma.res,
                              sigma_est=sigma.res, beta_est = beta.res, realLOD = exp(LOD))
          }
          
          res.simsum.sen[pos_res_sen:(pos_res_sen+n_par-1),] = cbind( rep(sen,n_par), 
                                                                      rep(methods[met],n_par), 
                                                                      c(solutions[[met]]$par, mcrps, marg_mean0, marg_mean1, marg_eff), #, marg_eff, marg_eff_lat, marg_mean0, marg_mean1), 
                                                                      c(SE, 1, 1, 1, 1), #, 0, 0, 0, 0), 
                                                                      par_name, rep(LOD_name,n_par), rep(dp_n[sen],n_par), FALSE) 
          pos_res_sen = pos_res_sen + n_par
        }else{ # One-part models
          
          if(methods[met] == "Substitute"){
            marg_eff = exp(gamma.res[2])
            marg_mean0 = exp(gamma.res[1] + sigma.res^2/2)
            marg_mean1 = exp(gamma.res[1] + gamma.res[2] + sigma.res^2/2)
            
            mcrps = mean.CRPS(y0, x0, m = 3, gamma_est = gamma.res,
                              sigma_est=sigma.res, realLOD = exp(LOD))
             
          }else if(methods[met] == "Tobit"){
            Z_0 = (LOD - gamma.res[1])/sigma.res
            Z_1 = (LOD - gamma.res[1] - gamma.res[2])/sigma.res
            
            marg_eff = (1 - pnorm(Z_1 - sigma.res)) / (1 - pnorm(Z_0 - sigma.res)) * exp(gamma.res[2])
            marg_mean0 = (1 - pnorm(Z_0 - sigma.res)) * exp(gamma.res[1] + sigma.res^2/2)
            marg_mean1 = (1 - pnorm(Z_1 - sigma.res)) * exp(gamma.res[1] + gamma.res[2] + sigma.res^2/2)
            
            mcrps = mean.CRPS(y0, x0, m = 4, gamma_est = gamma.res,
                              sigma_est=sigma.res, realLOD = exp(LOD))
          }
          
          
          res.simsum.sen[pos_res_sen:(pos_res_sen+n_par-1),] = cbind( rep(sen,n_par), 
                                                                      rep(methods[met],n_par), 
                                                                      c(solutions[[met]]$par, rep(NA,length(beta)), mcrps, marg_mean0, marg_mean1, marg_eff), #marg_eff, marg_eff_lat, marg_mean0, marg_mean1), 
                                                                      c(SE, rep(NA,length(beta)), 1, 1, 1, 1), 
                                                                      par_name, rep(LOD_name,n_par), rep(dp_n[sen],n_par), FALSE) 
          pos_res_sen = pos_res_sen + n_par
        }
      }
    }
  }
  # Calculating performance measures
  res.simsum.sen$theta = as.numeric(res.simsum.sen$theta)
  res.simsum.sen$se = as.numeric(res.simsum.sen$se)
  
  # Identify solutions with SE larger than 10 SD from average SE
  res.simsum.sen = cbind(res.simsum.sen, .dropbig = rep(FALSE, dim(res.simsum.sen)[1]))
  for(param in c("beta0", "beta1", "gamma0", "gamma1", "sigma")){
    res.simsum.sen[res.simsum.sen$par == param, ] = 
      dropbig(subset(res.simsum.sen, par == param), 
              estvarname = "theta", se = "se", methodvar = "method", by = c("LOD", "disc_prob"), 
              max = Inf, semax = 10, robust = FALSE) # .dropbig = TRUE for extreme outliers
  }
  data.all = rbind(data.all, res.simsum.sen) # Store all data
  
  # Remove extreme outliers
  if(sum(res.simsum.sen$.dropbig, na.rm = TRUE) > 0){
    res.simsum.sen[which(res.simsum.sen$.dropbig),]$theta = NA
    res.simsum.sen[which(res.simsum.sen$.dropbig),]$se = NA 
  }
  
  # Calculate performance measures (without extreme outliers)
  s.sen = multisimsum(data = res.simsum.sen, estvarname = "theta", par = "par", 
                      true = c(gamma0 = gamma[1], gamma1 = gamma[2],
                               beta0 = beta[1], beta1 = beta[2], sigma = sigma, 
                               mcrps = 0, marg_mean0 = marginal_mean0, marg_mean1 = marginal_mean1, marg_eff_obs = marginal_effect), 
                      se = "se", by = c("LOD", "disc_prob"), methodvar = "method", ref = 'TPIC', x = TRUE)
  
  summ.all = rbind(summ.all, summary(s.sen)$summ)
  res.list[[sen]] = s.sen
}

# Aesthetic Name Changes
{
names(summ.all)[names(summ.all) == "disc_prob"] = "Discrete"
names(summ.all)[names(summ.all) == "method"] = "Method"

summ.all$stat[summ.all$stat == "bias"] = "Bias"
summ.all$stat[summ.all$stat == "cover"] = "Coverage"
summ.all$stat[summ.all$stat == "power"] = "Power"
summ.all$stat[summ.all$stat == "empse"] = "EmpSE"
summ.all$stat[summ.all$stat == "modelse"] = "ModSE"
summ.all$stat[summ.all$stat == "thetamean"] = "Mean Value"
summ.all$stat[summ.all$stat == "relerror"] = "RelErrorSE"

summ.all$par = as.character(summ.all$par)
summ.all$par[summ.all$par == "marg_mean0"] = "Mean x = 0"
summ.all$par[summ.all$par == "marg_mean1"] = "Mean x = 1"
summ.all$par[summ.all$par == "marg_eff_obs"] = "Marginal Effect"

summ.all$Discrete = paste(summ.all$Discrete, "%")
#summ.all$Discrete = substr(summ.all$Discrete, 0,4)

names(data.all)[names(data.all) == "disc_prob"] = "Discrete"
data.all$Discrete = paste(data.all$Discrete, "%")
}

if(FALSE){
  data.all.old = data.all
  summ.all = data.frame(stat = character(), est = numeric(), mcse = numeric(), method = character(), 
                        LOD = character(), disc_prob = character(), beta1 = character(),
                        par = character(), lower = numeric(), upper = numeric())
  
  data.all = data.frame(dataset = integer(), method = character(), theta = numeric(), se = numeric(),
                        par = character(), LOD = character(), disc_prob = character(), .dropbig = logical())
  
  res.list = list()
  for(sen in 1:16){
    print(sen)
    res.simsum.sen = subset(data.all.old, dataset == sen)
    beta0 = beta0_vec[sen]
    beta = c(beta0, beta1)
    LOD = LOD_vec[sen]
    LOD_name = LOD_n[sen]
    Z_0 = (LOD - gamma0)/sigma
    C_0 = (1-pnorm(Z_0 - sigma))
    Z_1 = (LOD - gamma0 - gamma1)/sigma
    C_1 = (1-pnorm(Z_1 - sigma))
    marginal_effect = (1-pnorm(beta0 + beta1))/(1-pnorm(beta0)) * C_1 / C_0 * exp(gamma1)
    marginal_mean0 = (1-pnorm(beta0)) * (1-pnorm(Z_0 - sigma)) * exp(gamma0 + sigma^2/2)
    marginal_mean1 = (1-pnorm(beta0 + beta1)) * (1-pnorm(Z_1 - sigma)) * exp(gamma0 + gamma1 + sigma^2/2)
    
    # Drop big
    res.simsum.sen = cbind(res.simsum.sen, .dropbig = rep(FALSE, dim(res.simsum.sen)[1]))
    res.simsum.sen = cbind(res.simsum.sen, .dropbig = rep(FALSE, dim(res.simsum.sen)[1]))
    for(param in c("beta0", "beta1", "gamma0", "gamma1", "sigma")){
      res.simsum.sen[res.simsum.sen$par == param, ] = 
        dropbig(subset(res.simsum.sen, par == param), 
                estvarname = "theta", se = "se", methodvar = "method", by = c("LOD", "Discrete"), 
                max = Inf, semax = 10, robust = FALSE) # .dropbig = TRUE for extreme outliers
    }
    data.all = rbind(data.all, res.simsum.sen) # Store all data
    
    # Remove outliers
    if(sum(res.simsum.sen$.dropbig, na.rm = TRUE) > 0){
      res.simsum.sen[which(res.simsum.sen$.dropbig),]$theta = NA
      res.simsum.sen[which(res.simsum.sen$.dropbig),]$se = NA 
    }
    
    # Calculate performance measures
    s.sen = multisimsum(data = res.simsum.sen, estvarname = "theta", par = "par", 
                        true = c(gamma0 = gamma[1], gamma1 = gamma[2],
                                 beta0 = beta[1], beta1 = beta[2], sigma = sigma, 
                                 mcrps = 0, marg_mean0 = marginal_mean0, marg_mean1 = marginal_mean1, marg_eff_obs = marginal_effect), 
                        se = "se", by = c("LOD", "Discrete"), methodvar = "method", ref = 'TPIC', x = TRUE)
    
    summ.all = rbind(summ.all, summary(s.sen)$summ)
    res.list[[sen]] = s.sen 
  }
}

