
fit_model <-function(x, y, LOD, model, init, cont = "ln"){
  # x: Data, y: responses, LOD: Limit of Detection, 
  # model: Which model to fit, init: Initial values
  # cont: Which continuous distribution to use. 
  # (ln = lognormal, lsn = log-skew-normal)
  
  if(model == "ni"){ # Two-part 
    return(optim(par = init, fn = optim_likelihood_ni, x = x, y = y, 
                 LOD = LOD, cont = cont,  method = "BFGS", hessian = TRUE, 
                 control = list("fnscale"=-1)))
  }else if(model == "i"){ # Two-part w/ interval censoring
    return(optim(par = init, fn = optim_likelihood_i, x = x, y = y, 
                 LOD = LOD, cont = cont, method = "BFGS", hessian = TRUE, 
                 control = list("fnscale"=-1)))
  }else if(model == "pi"){ # Substitute model
    return(optim(par = init[1:ceiling((length(init)+1)/2)], 
                 fn = optim_likelihood_pi, x = x, y = y, LOD = LOD, cont = cont, 
                 method = "BFGS", hessian = TRUE, control = list("fnscale"=-1)))
  }else if(model == "t"){ # Tobit model
    return(optim(par = init[1:ceiling((length(init)+1)/2)], 
                 fn = optim_likelihood_t, x = x, y = y, LOD = LOD, cont = cont, 
                 method = "BFGS", hessian = TRUE, control = list("fnscale"=-1)))
  }else if(model == "mt"){ # Marginalized two-part model
    return(optim(par = init, fn = optim_likelihood_mt, x = x, y = y, 
                 LOD = LOD, cont = cont, method = "BFGS", hessian = TRUE, 
                 control = list("fnscale"=-1)))
  }
}


### TWO-PART W/O INTERVAL CENSORING:
log_likelihood_ni= function(x, y, beta, gamma, sigma, LOD, cont, delta = NA){
  # Log-likelihood of TP model
  
  censored = as.integer(is.na(y))
  likelihood = 0
  
  # Iterate over all observations
  for(i in 1:length(y)){
    y_i = y[i]
    x_i = x[i,]
    p = pnorm(sum(x_i * beta)) # probit link function 
    if(censored[i]){ # censored
      update = log(p)
    }else{ # not censored
      if(cont == "ln"){ # Lognormal continuous part
        update = log(1 - p) + dnorm(x = y_i, mean = sum(x_i * gamma), 
                                    sd = abs(sigma), log = TRUE)
      }else if(cont == "lsn"){ # Log-skew-normal continuous part
        update = log(1 - p) + dsn(x = y_i, xi = sum(x_i * gamma), 
                                  alpha = delta/abs(sigma), 
                                  omega = sqrt(sigma^2 + delta^2), log = TRUE)
      }
    }
    likelihood = likelihood + update
  }
  return(likelihood)
}

optim_likelihood_ni = function(par, x, y, LOD, cont){
  # Function to be maximized by optim for TP model 
  
  if(cont == "lsn"){ # Log-skew-normal continuous part
    delta = par[length(par)]
    par = par[-length(par)]
  }else{ # Lognormal continuous part
    delta = NA
  }
  gamma = par[1:floor(length(par)/2)]
  sigma = par[ceiling(length(par)/2)]
  beta = par[(ceiling(length(par)/2)+1):length(par)]
  return(log_likelihood_ni(x, y, beta, gamma, sigma, LOD, cont, delta)) 
}


### TWO-PART MODELT WITH INTERVAL CENSORING:
log_likelihood_i= function(x, y, beta, gamma, sigma, LOD, cont, delta = NA){
  # Log-likelihood of TPIC model
  
  censored = as.integer(is.na(y))
  likelihood = 0
  
  # Iterate over all observations
  for(i in 1:length(y)){
    y_i = y[i]
    x_i = x[i,]
    p = pnorm(sum(x_i * beta)) # probit link function 
    if(censored[i]){ # censored
      if(cont == "ln"){ # Lognormal continuous part
        update = log(p + (1 - p) * pnorm(q = LOD, mean = sum(x_i * gamma), 
                                         sd = abs(sigma)))
      }else if(cont == "lsn"){ # Log-skew-normal continuous part
        update = log(p + (1 - p) * psn(x = LOD, xi = sum(x_i * gamma), 
                                       alpha = delta/abs(sigma), 
                                       omega = sqrt(sigma^2 + delta^2)))
      }
      
    }else{ # not censored
      if(cont == "ln"){ # Lognormal continuous part
        update = log(1 - p) + dnorm(x = y_i, mean = sum(x_i * gamma), 
                                    sd = abs(sigma), log = TRUE) 
      }else if(cont == "lsn"){  # Log-skew-normal continuous part
        update = log(1 - p) + dsn(x = y_i, xi = sum(x_i * gamma), 
                                  alpha = delta/abs(sigma), 
                                  omega = sqrt(sigma^2 + delta^2), log = TRUE)
      }
    }
    likelihood = likelihood + update
  }
  return(likelihood)
}

optim_likelihood_i = function(par, x, y, LOD, cont){
  # Function to be maximized by optim for TPIC model 
  
  if(cont == "lsn"){ # Log-skew-normal continuous part
    delta = par[length(par)]
    par = par[-length(par)]
  }else{ # Lognormal continuous part
    delta = NA
  }
  gamma = par[1:floor(length(par)/2)]
  sigma = par[ceiling(length(par)/2)]
  beta = par[(ceiling(length(par)/2)+1):length(par)]
  return(log_likelihood_i(x, y, beta, gamma, sigma, LOD, cont, delta)) 
}


### SUBSTITUE MODEL
log_likelihood_pi= function(x, y, gamma, sigma, LOD, cont, delta){
  # Log-likelihood function of substitute model 
  
  y[is.na(y)] = LOD - log(2) # plugin LOD/2
  likelihood = 0
  
  # Iterate over all observations
  for(i in 1:length(y)){
    y_i = y[i]
    x_i = x[i,]
    if(cont == "ln"){ # Lognormal distribution 
      update = dnorm(x = y_i, mean = sum(x_i * gamma), 
                     sd = abs(sigma), log = TRUE)
    }else if(cont == "lsn"){ # Log-skew-normal distribution
      update = dsn(x = y_i, xi = sum(x_i * gamma), alpha = delta/abs(sigma), 
                   omega = sqrt(sigma^2 + delta^2), log = TRUE)
    }
    likelihood = likelihood + update
  }
  return(likelihood)
}

optim_likelihood_pi = function(par, x, y, LOD, cont){
  # Function to be maximized by optim for substitute model 
  
  if(cont == "lsn"){ # Log-skew-normal distribution
    delta = par[length(par)]
    par = par[-length(par)]
  }else{ # Log-normal distribution 
    delta = NA
  }
  gamma = par[1:(length(par)-1)]
  sigma = par[length(par)]
  return(log_likelihood_pi(x, y, gamma, sigma, LOD, cont, delta)) 
}


### TOBIT MODEL:
log_likelihood_t= function(x, y, gamma, sigma, LOD, cont, delta){
  # Log-likelihood function of substitute model 
  
  censored = as.integer(is.na(y))
  likelihood = 0
  
  # Iterate over all observations
  for(i in 1:length(y)){
    y_i = y[i]
    x_i = x[i,]
    if(censored[i]){ # censored
      if(cont == "ln"){ # Lognormal distribution 
        update = pnorm(q = LOD, mean = sum(x_i * gamma), 
                       sd = abs(sigma), log = TRUE)
      }else if(cont == "lsn"){ # Log-skew-normal distribution 
        update = log(psn(x = LOD, xi = sum(x_i * gamma), 
                         alpha = delta/abs(sigma), 
                         omega = sqrt(sigma^2 + delta^2)))
      }
    }else{ # not censored
      if(cont == "ln"){ # Lognormal distribution 
        update = dnorm(x = y_i, mean = sum(x_i * gamma), 
                       sd = abs(sigma), log = TRUE)
      }else if(cont == "lsn"){ # Log-skew-normal distribution 
        update = dsn(x = y_i, xi = sum(x_i * gamma), 
                     alpha = delta/abs(sigma), 
                     omega = sqrt(sigma^2 + delta^2), log = TRUE)
      }
    }
    likelihood = likelihood + update
  }
  return(likelihood)
}

optim_likelihood_t = function(par, x, y, LOD, cont){
  # Function to be maximized by optim for Tobit model 
  
  if(cont == "lsn"){ # Log-skew-normal distribution
    delta = par[length(par)]
    par = par[-length(par)]
  }else{ # Lognormal distribution 
    delta = NA
  }
  gamma = par[1:(length(par)-1)]
  sigma = par[length(par)]
  return(log_likelihood_t(x, y, gamma, sigma, LOD, cont, delta)) 
}
